import numpy as np

from gym.spaces import Box, Discrete, MultiDiscrete, MultiBinary, Tuple, Dict

def from_space(space,int_type,float_type):
    if isinstance(space,Discrete):
        return {"dtype": int_type,"shape": 1}
    elif isinstance(space,MultiDiscrete):
        return {"dtype": int_type,"shape": space.nvec.shape}
    elif isinstance(space,Box):
        return {"dtype": float_type,"shape": space.shape}
    elif isinstance(space,MultiBinary):
        return {"dtype": int_type, "shape": space.n}
    else:
        raise NotImplementedError(f"Error: Unknown Space {space}")

def create_env_dict(env,*,int_type = None,float_type = None):
    """
    Create `env_dict` from Open AI `gym.space` for `ReplayBuffer.__init__`

    Paremeters
    ----------
    env : gym.Env
        Environment
    int_type: np.dtype, optional
        Integer type. Default is `np.int32`
    float_type: np.dtype, optional
        Floating point type. Default is `np.float32`

    Returns
    -------
    env_dict : dict
        env_dict parameter for `ReplayBuffer` class.
    """

    int_type = int_type or np.int32
    float_type = float_type or np.float32

    env_dict = {"rew" : {"shape": 1, "dtype": float_type},
                "done": {"shape": 1, "dtype": float_type}}

    observation_space = env.observation_space
    action_space = env.action_space

    if isinstance(observation_space,Tuple):
        for i,s in enumerate(observation_space.spaces):
            env_dict[f"obs{i}"] = from_space(s,int_type,float_type)
            env_dict[f"next_obs{i}"] = from_space(s,int_type,float_type)
    elif isinstance(observation_space,Dict):
        for n, s in observation_space.spaces.items():
            env_dict[n] = from_space(s,int_type,float_type)
            env_dict[f"next_{n}"] = from_space(s,int_type,float_type)
    else:
        env_dict["obs"] = from_space(observation_space,int_type,float_type)
        env_dict["next_obs"] = from_space(observation_space,int_type,float_type)

    if isinstance(action_space,Tuple):
        for i,s in enumerate(action_space.spaces):
            env_dict[f"act{i}"] = from_space(s,int_type,float_type)
    elif isinstance(action_space,Dict):
        for n, s in action_space.spaces.items():
            env_dict[n] = from_space(s,int_type,float_type)
    else:
        env_dict["act"] = from_space(action_space,int_type,float_type)

    return env_dict

def create_before_add_func(env):
    """
    Create function to be used before `ReplayBuffer.add`

    Parameters
    ----------
    env : gym.Env
        Environment for before_func

    Returns
    -------
    before_add : callable
        Function to be used before `ReplayBuffer.add`
    """
    def no_convert(name,v):
        return {f"{name}": v}

    def convert_from_tuple(name,_tuple):
        return {f"{name}{i}": v for i,v in enumerate(_tuple)}

    def convert_from_dict(name,_dict):
        return {f"{name}_{key}":v for key,v in _dict.items()}


    observation_space = env.observation_space
    action_space = env.action_space


    if isinstance(observation_space,Tuple):
        obs_func = convert_from_tuple
    elif isinstance(observation_space,Dict):
        obs_func = convert_from_dict
    else:
        obs_func = no_convert

    if isinstance(action_space,Tuple):
        act_func = convert_from_tuple
    elif isinstance(action_space,Dict):
        act_func = convert_from_dict
    else:
        act_func = no_convert

    def before_add(obs,act,next_obs,rew,done):
        return {**obs_func("obs",obs),
                **act_func("act",act),
                **obs_func("next_obs",next_obs),
                "rew": rew,
                "done": done}

    return before_add



import os
import platform
import warnings
from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext

debug = os.getenv('DEBUG_CPPRB')

requires = ["numpy"]
setup_requires = ["numpy<1.20"]

rb_source = "cpprb/PyReplayBuffer"
cpp_ext = ".cpp"
pyx_ext = ".pyx"

extras = {
    'gym': ["matplotlib", "pyvirtualdisplay"],
    'api': ["sphinx","sphinx_rtd_theme","sphinx-automodapi"],
    'dev': ["coverage","cython","gym[box2d]","twine","unittest-xml-reporting","wheel"]
}
all_deps = []
for group_name in extras:
    all_deps += extras[group_name]
extras['all'] = all_deps

# Set compiler flags depending on platform
if platform.system() == 'Windows':
    extra_compile_args = ["/std:c++17"]
    extra_link_args = None
    if debug:
        extra_compile_args.append('/DCYTHON_TRACE_NOGIL=1')
else:
    extra_compile_args = ["-std=c++17","-march=native"]
    extra_link_args = ["-std=c++17", "-pthread"]
    if debug:
        extra_compile_args.append('-DCYTHON_TRACE_NOGIL=1')

# Check cythonize or not
cpp_file = rb_source + cpp_ext
pyx_file = rb_source + pyx_ext
use_cython = (not os.path.exists(cpp_file)
              or (os.path.exists(pyx_file)
                  and (os.path.getmtime(cpp_file) < os.path.getmtime(pyx_file))))
if use_cython:
    suffix = pyx_ext
    setup_requires.extend(["cython>=0.29"])
    compiler_directives = {'language_level': "3"}

    if debug:
        compiler_directives['linetrace'] = True
else:
    suffix = cpp_ext

# Set ext_module
ext = [["cpprb","PyReplayBuffer"],
       ["cpprb","VectorWrapper"]]

ext_modules = [Extension(".".join(e),
                         sources=["/".join(e) + suffix],
                         extra_compile_args=extra_compile_args,
                         extra_link_args=extra_link_args,
                         language="c++") for e in ext]

class LazyImportBuildExtCommand(build_ext):
    """
    build_ext command class for lazy numpy and cython import
    """
    def run(self):
        import numpy as np

        self.include_dirs.append(np.get_include())
        build_ext.run(self)

    def finalize_options(self):
        if use_cython:
            from Cython.Build import cythonize
            self.distribution.ext_modules = cythonize(self.distribution.ext_modules,
                                                      compiler_directives=compiler_directives,
                                                      include_path=["."],
                                                      annotate=True)
        super().finalize_options()


description = "ReplayBuffer for Reinforcement Learning written by C++ and Cython"
README = os.path.join(os.path.abspath(os.path.dirname(__file__)),'README.md')
if os.path.exists(README):
    with open(README,encoding='utf-8') as f:
        long_description = f.read()
    long_description_content_type='text/markdown'
else:
    warnings.warn("No README.md")
    long_description =  description
    long_description_content_type='text/plain'

setup(name="cpprb",
      author="Yamada Hiroyuki",
      description=description,
      version="10.4.1",
      install_requires=requires,
      setup_requires=setup_requires,
      extras_require=extras,
      cmdclass={'build_ext': LazyImportBuildExtCommand},
      url="https://ymd_h.gitlab.io/cpprb/",
      ext_modules=ext_modules,
      include_dirs=["cpprb"],
      packages=["cpprb"],
      classifiers=["Programming Language :: Python",
                   "Programming Language :: Python :: 3",
                   "License :: OSI Approved :: MIT License",
                   "Operating System :: OS Independent",
                   "Development Status :: 4 - Beta",
                   "Intended Audience :: Developers",
                   "Intended Audience :: Science/Research",
                   "Topic :: Scientific/Engineering",
                   "Topic :: Scientific/Engineering :: Artificial Intelligence",
                   "Topic :: Software Development :: Libraries"],
      long_description=long_description,
      long_description_content_type=long_description_content_type)



import numpy as np

class LaBER:
    def __init__(self, batch_size: int, m: int = 4, *, eps: float = 1e-6):
        """
        Initialize LaBER (sub-)class

        Ref: https://arxiv.org/abs/2110.01528

        Parameters
        ----------
        batch_size : int
            Batch size for neural network
        m : int, optional
            Multiplication factor. `m * batch_size` transitions will be passed.
            Default value is `4`.
        eps : float, option
            Small positive values to avoid 0 priority. Default value is `1e-6`.
        """
        self.rng = np.random.default_rng()

        self.batch_size = int(batch_size)
        if self.batch_size <= 0:
            raise ValueError("`batch_size` must be positive integer.")

        if m <= 0:
            raise ValueError("`m` must be positive integer")

        self.idx = np.arange(int(self.batch_size * m))

        self.eps = float(eps)
        if self.eps < 0:
            raise ValueError("`eps` must be non negative")

    def __call__(self, *, priorities, **kwargs):
        """
        Sub-sample from large batch

        Parameters
        ----------
        priorities : array-like of float
            Surrogate priorities.
        **kwargs : key-value
            Large batch sampled from `ReplayBuffer`.
        """
        p = np.asarray(priorities) + self.eps
        if p.shape != self.idx.shape:
            raise ValueError("`priorities` size must be `batch_size * m`")

        p = p / p.sum()

        _idx = self.rng.choice(self.idx, self.batch_size, p=p)

        if kwargs is None:
            kwargs = {}
        else:
            kwargs = {k: v[_idx] for k, v in kwargs.items()}

        kwargs["weights"] = self._normalize_weight(p, _idx)
        kwargs["indexes"] = _idx

        return kwargs

    def _normalize_weight(self, p, _idx):
        raise NotImplementedError


class LaBERmean(LaBER):
    def _normalize_weight(self, p, _idx):
        return p.mean() / p[_idx]


class LaBERlazy(LaBER):
    def _normalize_weight(self, p, _idx):
        return 1.0 / p[_idx]


class LaBERmax(LaBER):
    def _normalize_weight(self, p, _idx):
        p_idx = 1.0 / p[_idx]
        return p_idx / p_idx.max()



from .PyReplayBuffer import (ReplayBuffer,PrioritizedReplayBuffer,
                             MPReplayBuffer, MPPrioritizedReplayBuffer,
                             SelectiveReplayBuffer, ReverseReplayBuffer)

from .LaBER import LaBERmean, LaBERlazy, LaBERmax

from .PyReplayBuffer import create_buffer, train

try:
    from .util import create_env_dict, create_before_add_func
except ImportError:
    # If gym is not installed, util functions are not defined.
    pass
